import ast
import sys
from pathlib import Path
from .pifm_errors import raise_pifm_error, format_code_block


class TemplateValidator(ast.NodeVisitor):
    def __init__(self, file_path, template_path, *args, **kwargs):
        self.file_path = Path(file_path)
        self.template_path = Path(template_path)

        self.file_text = self.file_path.read_text()
        self.template_text = self.template_path.read_text()

        self.file_ast = ast.parse(self.file_text)
        self.template_ast = ast.parse(self.template_text)

        self.valid_ast = self.template_ast.body

        self.node_parent = None

        super().__init__(*args, **kwargs)

    def validate(self):
        self.visit(self.file_ast)

    def generic_visit(self, node):
        old_node = self.node_parent
        self.node_parent = node
        super().generic_visit(node)
        self.node_parent = old_node

    def raise_error_for_node(self, node, error_note):
        code_block = format_code_block(
            "The location where the error occured is shown below",
            self.file_text,
            self.file_path,
            node.lineno,
            node.col_offset,
            node.end_lineno,
            node.end_col_offset
        )


        raise_pifm_error(error_note, blocks=[code_block])

    @staticmethod
    def are_ast_trees_equal(tree1, tree2):
        if isinstance(tree1, list):
            tree1 = [ast.dump(node, include_attributes=False) for node in tree1]
        else:
            tree1 = ast.dump(tree1, include_attributes=False)

        if isinstance(tree2, list):
            tree2 = [ast.dump(node, include_attributes=False) for node in tree2]
        else:
            tree2 = ast.dump(tree2, include_attributes=False)

        return tree1 == tree2

    def parent_starts_with_functions_like(self, comparison_node):
        parent = self.node_parent
        for node in parent.body:
            if self.are_ast_trees_equal(node, comparison_node):
                return True

            if not isinstance(node, type(comparison_node)):
                return False

            if node.value.func.id != comparison_node.value.func.id:
                return False

        raise ValueError("It should not be possible for comparison_node to not be in the parent.")


    def visit_FunctionDef(self, node):
        top_level_functions = dict([
            (ast_node.name, ast_node)
            for ast_node in self.valid_ast
            if type(ast_node) == ast.FunctionDef
        ])
        if node.name not in top_level_functions:
            return

        current_function = top_level_functions[node.name]

        if not self.are_ast_trees_equal(node.args, current_function.args):
            self.raise_error_for_node(
                node,
                f"The function {node.name}'s arguments were not what was expected.\n"
                f"Check the arguments of that function are the same as the starter code."
            )

        containing_valid_ast = self.valid_ast
        self.valid_ast = current_function.body
        self.generic_visit(node)

        if self.valid_ast:
            self.raise_error_for_node(
                node,
                f"You are missing starter code that pifm expected you would have.\n"
                f"Make sure your code includes everything from the starter code."
            )
        else:
            containing_valid_ast.remove(current_function)
            self.valid_ast = containing_valid_ast

    def visit_Expr(self, node):
        if not self.valid_ast:
            # If we don't have any ast to validate, we can return.
            return

        if type(node.value) != ast.Call:
            # If we're not calling a function, we're all good.
            return

        if isinstance(self.node_parent, ast.Module):
            # We should ignore exprs that are in the root module
            return

        function_name = node.value.func.id

        valid_ast_node = self.valid_ast[0]
        valid_ast_node_name = valid_ast_node.value.func.id

        if valid_ast_node_name != function_name:
            return

        if not self.are_ast_trees_equal(node.value.args, valid_ast_node.value.args):
            self.raise_error_for_node(
                node,
                f"Pifm expected you to have a {valid_ast_node_name} call here that matched the starter code.\n"
                f"What you provided did not match the starter code."
            )

        if valid_ast_node_name == 'requires':
            if not isinstance(self.node_parent, ast.FunctionDef) or not self.parent_starts_with_functions_like(node):
                self.raise_error_for_node(
                    node,
                    f"You must always have your `requires` function at the start of a function's body."
                )

        if valid_ast_node_name == 'holds_invariant':
            if not type(self.node_parent) in [ast.While, ast.For] or not self.parent_starts_with_functions_like(node):
                self.raise_error_for_node(
                    node,
                    f"You must always have your `holds_invariant` function at the start of a function's body."
                )

        self.valid_ast.remove(valid_ast_node)



# Welcome to pifm (python in-formal methods), a wrapper for COMP6721
#
# The actions of this file break down into five parts:
# A: Introucing some new concepts to python (`requires`, `ensures`,
#    `holds_invariant`); and checking the importing code
#    uses them appropriately.
#
# B: Only allowing some python features in the calling file to ensure
#    students use a "pseudo-code" like function. Status: DONE.
#
# C: For named functions (i.e. ones the student has explicitly been asked to write);
#    we can use hypothesis testing to find assertion failures; by removing our
#    invariants. [EXPERIMENTALLY] We can even explore using students' `requires` functions as
#    predicates for hypothesis testing.
#
# D: Providing a simple utility that makes iFM a bit easier:
#     * Array
#
# E: [NOT DONE YET] using mypy to make sure we have types for students functions.

FILE_AST = None

import ast
import inspect
import os
import pathlib
import sys

from collections import defaultdict

from .astpp import dump as astpp_dump
from .array import Array
from .template_ast_validator import TemplateValidator
from .usage_ast_validator import UsageValidator

class PifmAssertionError(AssertionError):
    def __init__(self, message, info_tuple=(None, None)):
        self.message = message
        self.name, self.calling_function = info_tuple

    def as_tuple(self):
        return (self.name, self.calling_function)

    def __repr__(self):
        return f"PifmAssertionError<name={self.name}, calling_function={self.calling_function}>"

_ignored_assertions = []
_assertion_stats = defaultdict(lambda: 0)

def _ignore_assertion(assertion_tuple):
    print(_ignored_assertions)
    _ignored_assertions.append(assertion_tuple)

def _real_assert(assertion_name, arg):
    function_name = inspect.stack()[1].function
    assertion_tuple = (assertion_name, function_name)
    _assertion_stats[assertion_tuple] += 1
    if not arg and assertion_tuple not in _ignored_assertions:
        raise PifmAssertionError(
            f"Assertion {assertion_name} failed in function {function_name}.",
            info_tuple=assertion_tuple
        )

requires = _real_assert
ensures = _real_assert
holds_invariant = _real_assert


def pifm_initialise(template_name):
    """
    This function is  used to ensure elements A and B. It validates the entire
    AST of the called file.
    """
    calling_module_path = inspect.stack()[1].filename

    UsageValidator(calling_module_path).validate()

    current_path = pathlib.Path(__file__).parent.resolve()

    env_template_folder = os.getenv("PIFM_TEMPLATES_FOLDER")
    if env_template_folder:
        template_folder = pathlib.Path(env_template_folder)
    else:
        template_folder = current_path / "templates"

    template_path = template_folder / f"{template_name}.py"
    if not template_path.exists():
        print(f"pifm could not find the template {template_name}", file=sys.stderr)
        print(f"\tCheck your call to `pifm_initialise()`", file=sys.stderr)
        sys.exit(1)

    TemplateValidator(calling_module_path, template_path).validate()



class Array(list):
    """
    `Array` is a python list, with a few handy features:
    * if `array = Array()`, `array[len(array)]` can be assigned to (i.e. append to the array)
    * slices of Array are also Array.
    """

    def __add__(self, rhs):
        return Array(list.__add__(self, rhs))

    def __getslice__(self,i,j):
        return Array(list.__getslice__(self, i, j))

    def __mul__(self,other):
        return Array(list.__mul__(self,other))

    def __getitem__(self, item):
        result = list.__getitem__(self, item)
        if type([]) == type(result):
            return Array(result)
        else:
            return result

    def __setitem__(self, key, new_value):
        if key == len(self):
            self.append(new_value)
        else:
            super().__setitem__(key, new_value)



import ast
import sys
from pathlib import Path
from .node_rules import get_rule_for_node, Rules
from .pifm_errors import raise_pifm_error, format_code_block

class UsageValidator(ast.NodeVisitor):
    def __init__(self, file_path, *args, **kwargs):
        self.file_path = Path(file_path)
        self.file_text = self.file_path.read_text()
        self.file_ast = ast.parse(self.file_text)
        super().__init__(*args, **kwargs)

    def validate(self):
        self.generic_visit(self.file_ast)

    def raise_error_for_node(self, node, error_note):
        code_block = format_code_block(
            "The location where the error occured is shown below",
            self.file_text,
            self.file_path,
            node.lineno,
            node.col_offset,
            node.end_lineno,
            node.end_col_offset
        )

        raise_pifm_error(error_note, blocks=[code_block])

    def generic_visit(self, node):
        rule_for_node = get_rule_for_node(node)
        if rule_for_node == Rules.DISALLOWED:
            self.raise_error_for_node(node, f"pifm does not allow the use of {type(node).__name__}.")
        elif rule_for_node == Rules.UNKNOWN:
            self.raise_error_for_node(
                node,
                f"pifm does not know what a {type(node).__name__} node is.\n"
                f"It is likely this is a bug with pifm, please let the maintainers know."
            )
        super().generic_visit(node)



import sys

def raise_pifm_error(error_text, blocks=[]):
    print("Your code is not valid pifm code!", file=sys.stderr)
    print('\t' + error_text.replace('\n', '\n\t') + '\n', file=sys.stderr)

    for block in blocks:
        print(block, file=sys.stderr)

    sys.exit(0)

DEFAULT_CONTEXT = 3 # lines of context to show

def format_code_block(hint_text, code_text, file_name, real_start_line, real_col_offset, real_end_line, real_end_col_offset):
    formatted_text = f"Note: {hint_text} (at {file_name}:{real_start_line}:{real_col_offset} to {real_end_line}:{real_end_col_offset})\n"

    code_lines = code_text.split('\n')
    index_start_line = real_start_line - 1
    display_start_line = max(0, index_start_line - DEFAULT_CONTEXT)
    index_end_line = real_end_line - 1
    display_end_line = min(len(code_lines) - 1, index_end_line + DEFAULT_CONTEXT)

    for display_line_number in range(display_start_line, display_end_line + 1):
        formatted_text += f"{display_line_number + 1:4} "
        if display_line_number in range(index_start_line, index_end_line + 1):
            formatted_text += "> "
        else:
            formatted_text += "  "

        formatted_text += code_lines[display_line_number] + "\n"

    return formatted_text



import ast

"""
Stolen from the python documentation: https://docs.python.org/3.7/library/ast.html
We then copy this in below.

mod = Module(stmt* body)
        | Interactive(stmt* body)
        | Expression(expr body)

        -- not really an actual node but useful in Jython's typesystem.
        | Suite(stmt* body)

    stmt = FunctionDef(identifier name, arguments args,
                       stmt* body, expr* decorator_list, expr? returns)
          | AsyncFunctionDef(identifier name, arguments args,
                             stmt* body, expr* decorator_list, expr? returns)

          | ClassDef(identifier name,
             expr* bases,
             keyword* keywords,
             stmt* body,
             expr* decorator_list)
          | Return(expr? value)

          | Delete(expr* targets)
          | Assign(expr* targets, expr value)
          | AugAssign(expr target, operator op, expr value)
          -- 'simple' indicates that we annotate simple name without parens
          | AnnAssign(expr target, expr annotation, expr? value, int simple)

          -- use 'orelse' because else is a keyword in target languages
          | For(expr target, expr iter, stmt* body, stmt* orelse)
          | AsyncFor(expr target, expr iter, stmt* body, stmt* orelse)
          | While(expr test, stmt* body, stmt* orelse)
          | If(expr test, stmt* body, stmt* orelse)
          | With(withitem* items, stmt* body)
          | AsyncWith(withitem* items, stmt* body)

          | Raise(expr? exc, expr? cause)
          | Try(stmt* body, excepthandler* handlers, stmt* orelse, stmt* finalbody)
          | Assert(expr test, expr? msg)

          | Import(alias* names)
          | ImportFrom(identifier? module, alias* names, int? level)

          | Global(identifier* names)
          | Nonlocal(identifier* names)
          | Expr(expr value)
          | Pass | Break | Continue

          -- XXX Jython will be different
          -- col_offset is the byte offset in the utf8 string the parser uses
          attributes (int lineno, int col_offset)

          -- BoolOp() can use left & right?
    expr = BoolOp(boolop op, expr* values)
         | BinOp(expr left, operator op, expr right)
         | UnaryOp(unaryop op, expr operand)
         | Lambda(arguments args, expr body)
         | IfExp(expr test, expr body, expr orelse)
         | Dict(expr* keys, expr* values)
         | Set(expr* elts)
         | ListComp(expr elt, comprehension* generators)
         | SetComp(expr elt, comprehension* generators)
         | DictComp(expr key, expr value, comprehension* generators)
         | GeneratorExp(expr elt, comprehension* generators)
         -- the grammar constrains where yield expressions can occur
         | Await(expr value)
         | Yield(expr? value)
         | YieldFrom(expr value)
         -- need sequences for compare to distinguish between
         -- x < 4 < 3 and (x < 4) < 3
         | Compare(expr left, cmpop* ops, expr* comparators)
         | Call(expr func, expr* args, keyword* keywords)
         | Num(object n) -- a number as a PyObject.
         | Str(string s) -- need to specify raw, unicode, etc?
         | FormattedValue(expr value, int? conversion, expr? format_spec)
         | JoinedStr(expr* values)
         | Bytes(bytes s)
         | NameConstant(singleton value)
         | Ellipsis
         | Constant(constant value)

         -- the following expression can appear in assignment context
         | Attribute(expr value, identifier attr, expr_context ctx)
         | Subscript(expr value, slice slice, expr_context ctx)
         | Starred(expr value, expr_context ctx)
         | Name(identifier id, expr_context ctx)
         | List(expr* elts, expr_context ctx)
         | Tuple(expr* elts, expr_context ctx)

          -- col_offset is the byte offset in the utf8 string the parser uses
          attributes (int lineno, int col_offset)

    expr_context = Load | Store | Del | AugLoad | AugStore | Param

    slice = Slice(expr? lower, expr? upper, expr? step)
          | ExtSlice(slice* dims)
          | Index(expr value)

    boolop = And | Or

    operator = Add | Sub | Mult | MatMult | Div | Mod | Pow | LShift
                 | RShift | BitOr | BitXor | BitAnd | FloorDiv

    unaryop = Invert | Not | UAdd | USub

    cmpop = Eq | NotEq | Lt | LtE | Gt | GtE | Is | IsNot | In | NotIn

    # ... The rest is included in other things, and therefore
    # doesn't require checking.
}
"""

from enum import Enum

class Rules(Enum):
    DISALLOWED = 0
    ALLOWED = 1
    UNKNOWN = 2

# This is manually crafted to rules that seemed sensible to me. ~tfpk
NODE_RULES = {
    # mod:
    ast.Module: Rules.ALLOWED,
    ast.Interactive: Rules.ALLOWED,
    ast.Expression: Rules.DISALLOWED,

    # stmt:
    ast.FunctionDef: Rules.ALLOWED,
    ast.AsyncFunctionDef: Rules.DISALLOWED,
    ast.ClassDef: Rules.DISALLOWED,
    ast.Return: Rules.ALLOWED,

    ast.Delete: Rules.ALLOWED,
    ast.Assign: Rules.ALLOWED,
    ast.AugAssign: Rules.ALLOWED,
    ast.AnnAssign: Rules.ALLOWED,

    ast.For: Rules.ALLOWED,
    ast.AsyncFor: Rules.DISALLOWED,
    ast.While: Rules.ALLOWED,
    ast.If: Rules.ALLOWED,
    ast.With: Rules.DISALLOWED,
    ast.AsyncWith: Rules.DISALLOWED,

    ast.Raise: Rules.DISALLOWED,
    ast.Try: Rules.DISALLOWED,
    ast.Assert: Rules.DISALLOWED,

    ast.Import: Rules.ALLOWED,
    ast.ImportFrom: Rules.ALLOWED,

    ast.Global: Rules.DISALLOWED,
    ast.Nonlocal: Rules.DISALLOWED,
    ast.Expr: Rules.ALLOWED,
    ast.Pass: Rules.ALLOWED,
    ast.Break: Rules.ALLOWED,
    ast.Continue: Rules.ALLOWED,

    # expr:
    ast.BoolOp: Rules.ALLOWED,
    ast.BinOp: Rules.ALLOWED,
    ast.UnaryOp: Rules.ALLOWED,
    ast.Lambda: Rules.DISALLOWED,
    ast.IfExp: Rules.ALLOWED,
    ast.Dict: Rules.ALLOWED,
    ast.Set: Rules.ALLOWED,
    ast.ListComp: Rules.DISALLOWED,
    ast.SetComp: Rules.DISALLOWED,
    ast.DictComp: Rules.DISALLOWED,
    ast.GeneratorExp: Rules.DISALLOWED,
    ast.Await: Rules.DISALLOWED,
    ast.Yield: Rules.DISALLOWED,
    ast.YieldFrom: Rules.DISALLOWED,
    ast.Compare: Rules.ALLOWED,
    ast.Call: Rules.ALLOWED,
    ast.Num: Rules.ALLOWED,
    ast.Str: Rules.ALLOWED,
    ast.FormattedValue: Rules.ALLOWED,
    ast.JoinedStr: Rules.ALLOWED,
    ast.Bytes: Rules.ALLOWED,
    ast.NameConstant: Rules.ALLOWED,
    ast.Ellipsis: Rules.ALLOWED,
    ast.Constant: Rules.ALLOWED,

    ast.Attribute: Rules.ALLOWED,
    ast.Subscript: Rules.ALLOWED,
    ast.Starred: Rules.ALLOWED,
    ast.Name: Rules.ALLOWED,
    ast.List: Rules.ALLOWED,
    ast.Tuple: Rules.ALLOWED,

    ast.Load: Rules.ALLOWED,
    ast.Store: Rules.ALLOWED,
    ast.Del: Rules.ALLOWED,
    ast.AugLoad: Rules.ALLOWED,
    ast.AugStore: Rules.ALLOWED,
    ast.Param: Rules.ALLOWED,

    ast.Slice: Rules.ALLOWED,
    ast.ExtSlice: Rules.ALLOWED,
    ast.Index: Rules.ALLOWED,

    ast.And: Rules.ALLOWED,
    ast.Or: Rules.ALLOWED,

    ast.Add: Rules.ALLOWED,
    ast.Sub: Rules.ALLOWED,
    ast.Mult: Rules.ALLOWED,
    ast.MatMult: Rules.ALLOWED,
    ast.Div: Rules.ALLOWED,
    ast.Mod: Rules.ALLOWED,
    ast.Pow: Rules.ALLOWED,
    ast.LShift: Rules.ALLOWED,
    ast.RShift: Rules.ALLOWED,
    ast.BitOr: Rules.ALLOWED,
    ast.BitXor: Rules.ALLOWED,
    ast.BitAnd: Rules.ALLOWED,
    ast.FloorDiv: Rules.ALLOWED,

    ast.Invert: Rules.ALLOWED,
    ast.Not: Rules.ALLOWED,
    ast.UAdd: Rules.ALLOWED,
    ast.USub: Rules.ALLOWED,

    ast.Eq: Rules.ALLOWED,
    ast.NotEq: Rules.ALLOWED,
    ast.Lt: Rules.ALLOWED,
    ast.LtE: Rules.ALLOWED,
    ast.Gt: Rules.ALLOWED,
    ast.GtE: Rules.ALLOWED,
    ast.Is: Rules.ALLOWED,
    ast.IsNot: Rules.ALLOWED,
    ast.In: Rules.ALLOWED,
    ast.NotIn: Rules.ALLOWED,

    ast.comprehension: Rules.ALLOWED,
    ast.excepthandler: Rules.ALLOWED,
    ast.arguments: Rules.ALLOWED,
    ast.arg: Rules.ALLOWED,
    ast.keyword: Rules.ALLOWED,
    ast.alias: Rules.ALLOWED,
    ast.withitem: Rules.ALLOWED,
}

def get_rule_for_node(node):
    return NODE_RULES.get(type(node), Rules.UNKNOWN)
